@@ -80,7 +80,9 @@ def mnemonic_is_valid(mnemonic: str, wordlist=WORDLIST):
 def mnemonic_to_seed(mnemonic: str, password: str = "", wordlist=WORDLIST):
     # first we try to convert mnemonic to bytes
     # and raise a correct error if it is invalid
-    mnemonic_to_bytes(mnemonic, wordlist=wordlist)
+    # If wordlist is None - don't check mnemonic.
+    if wordlist is not None:
+        mnemonic_to_bytes(mnemonic, wordlist=wordlist)
     return hashlib.pbkdf2_hmac(
         "sha512",
         mnemonic.encode("utf-8"),
@@ -99,18 +101,18 @@ def _extract_index(bits, b, n):
     return value
 
 
-def mnemonic_from_bytes(b, wordlist=WORDLIST):
-    if len(b) % 4 != 0:
+def mnemonic_from_bytes(entropy, wordlist=WORDLIST):
+    if len(entropy) % 4 != 0:
         raise ValueError("Byte array should be multiple of 4 long (16, 20, ..., 32)")
-    total_bits = len(b) * 8
+    total_bits = len(entropy) * 8
     checksum_bits = total_bits // 32
     total_mnemonics = (total_bits + checksum_bits) // 11
     # no need to truncate checksum - we already know total_mnemonics
-    checksum = bytearray(hashlib.sha256(b).digest())
-    b += checksum
+    checksum = bytearray(hashlib.sha256(entropy).digest())
+    entropy += checksum
     mnemonic = []
     for i in range(0, total_mnemonics):
-        idx = _extract_index(11, b, i)
+        idx = _extract_index(11, entropy, i)
         mnemonic.append(wordlist[idx])
     return " ".join(mnemonic)
 
