@@ -7,12 +7,17 @@ from . import hashes
 from .script import Script, Witness
 from . import script
 from .base import EmbitBase, EmbitError
+
 from binascii import b2a_base64, a2b_base64, hexlify, unhexlify
 from io import BytesIO
 
 class PSBTError(EmbitError):
     pass
 
+class CompressMode:
+    KEEP_ALL = 0
+    CLEAR_ALL = 1
+    PARTIAL = 2
 
 def ser_string(stream, s: bytes) -> int:
     return stream.write(compact.to_bytes(len(s))) + stream.write(s)
@@ -109,7 +114,7 @@ class InputScope(PSBTScope):
     TX_CLS = Transaction
     TXOUT_CLS = TransactionOutput
 
-    def __init__(self, unknown: dict = {}, vin=None, compress=False):
+    def __init__(self, unknown: dict = {}, vin=None, compress=CompressMode.KEEP_ALL):
         self.compress = compress
         self.txid = None
         self.vout = None
@@ -129,19 +134,32 @@ class InputScope(PSBTScope):
         self.redeem_script = None
         self.witness_script = None
         self.bip32_derivations = OrderedDict()
+
+        # tuples of ([leaf_hashes], DerivationPath)
+        self.taproot_bip32_derivations = OrderedDict()
+        self.taproot_internal_key = None
+
         self.final_scriptsig = None
         self.final_scriptwitness = None
         self.parse_unknowns()
 
-    def clear_metadata(self):
+    def clear_metadata(self, compress=CompressMode.CLEAR_ALL):
         """Removes metadata like derivations, utxos etc except final or partial sigs"""
+        if compress == CompressMode.KEEP_ALL:
+            return
         self.unknown = {}
-        self.non_witness_utxo = None
-        self.witness_utxo = None
-        self.sighash_type = None
-        self.redeem_script = None
-        self.witness_script = None
+        if compress == CompressMode.CLEAR_ALL:
+            self.non_witness_utxo = None
+            self.witness_utxo = None
+            self.sighash_type = None
+            self.redeem_script = None
+            self.witness_script = None
+        else:
+            if self.witness_utxo is not None:
+                self.non_witness_utxo = None
         self.bip32_derivations = OrderedDict()
+        self.taproot_bip32_derivations = OrderedDict()
+        self.taproot_internal_key = None
 
     def update(self, other):
         self.txid = other.txid or self.txid
@@ -156,6 +174,8 @@ class InputScope(PSBTScope):
         self.redeem_script = other.redeem_script or self.redeem_script
         self.witness_script = other.witness_script or self.witness_script
         self.bip32_derivations.update(other.bip32_derivations)
+        self.taproot_bip32_derivations.update(other.taproot_bip32_derivations)
+        self.taproot_internal_key = other.taproot_internal_key
         self.final_scriptsig = other.final_scriptsig or self.final_scriptsig
         self.final_scriptwitness = other.final_scriptwitness or self.final_scriptwitness
 
@@ -219,7 +239,9 @@ class InputScope(PSBTScope):
                     tx = self.TX_CLS.read_from(stream)
                     self.non_witness_utxo = tx
             return
+
         v = read_string(stream)
+
         # witness utxo
         if k[0] == 0x01:
             if len(k) != 1:
@@ -264,13 +286,15 @@ class InputScope(PSBTScope):
                 self.witness_script = Script(v)
             else:
                 raise PSBTError("Duplicated witness script")
-        # bip32 derivation
+
+        # PSBT_IN_BIP32_DERIVATION
         elif k[0] == 0x06:
             pub = ec.PublicKey.parse(k[1:])
             if pub in self.bip32_derivations:
                 raise PSBTError("Duplicated derivation path")
             else:
                 self.bip32_derivations[pub] = DerivationPath.parse(v)
+
         # final scriptsig
         elif k[0] == 0x07:
             # we don't need this key for signing
@@ -293,12 +317,30 @@ class InputScope(PSBTScope):
                 self.final_scriptwitness = Witness.parse(v)
             else:
                 raise PSBTError("Duplicated final scriptwitness")
+        
         elif k == b"\x0e":
             self.txid = bytes(reversed(v))
         elif k == b"\x0f":
             self.vout = int.from_bytes(v, 'little')
         elif k == b"\x10":
             self.sequence = int.from_bytes(v, 'little')
+
+        # PSBT_IN_TAP_BIP32_DERIVATION
+        elif k[0] == 0x16:
+            pub = ec.PublicKey.from_xonly(k[1:])
+            if pub not in self.taproot_bip32_derivations:
+                b = BytesIO(v)
+                num_leaf_hashes = compact.read_from(b)
+                leaf_hashes = [b.read(32) for i in range(num_leaf_hashes)]
+                if not all([len(leaf)==32 for leaf in leaf_hashes]):
+                    raise PSBTError("Invalid length of taproot leaf hashes")
+                der = DerivationPath.read_from(b)
+                self.taproot_bip32_derivations[pub] = (leaf_hashes, der)
+        
+        # PSBT_IN_TAP_INTERNAL_KEY
+        elif k[0] == 0x17:
+            self.taproot_internal_key = ec.PublicKey.from_xonly(v)
+
         else:
             if k in self.unknown:
                 raise PSBTError("Duplicated key")
@@ -333,6 +375,7 @@ class InputScope(PSBTScope):
         if self.final_scriptwitness is not None:
             r += stream.write(b"\x01\x08")
             r += ser_string(stream, self.final_scriptwitness.serialize())
+
         if version == 2:
             if self.txid is not None:
                 r += ser_string(stream, b"\x0e")
@@ -343,6 +386,18 @@ class InputScope(PSBTScope):
             if self.sequence is not None:
                 r += ser_string(stream, b"\x10")
                 r += ser_string(stream, self.sequence.to_bytes(4, 'little'))
+
+        # PSBT_IN_TAP_BIP32_DERIVATION
+        for pub in self.taproot_bip32_derivations:
+            r += ser_string(stream, b"\x16" + pub.xonly())
+            leaf_hashes, derivation = self.taproot_bip32_derivations[pub]
+            r += ser_string(stream, compact.to_bytes(len(leaf_hashes)) + b"".join(leaf_hashes) + derivation.serialize())
+
+        # PSBT_IN_TAP_INTERNAL_KEY
+        if self.taproot_internal_key is not None:
+            r += ser_string(stream, b"\x17")
+            r += ser_string(stream, self.taproot_internal_key.xonly())
+
         # unknown
         for key in self.unknown:
             r += ser_string(stream, key)
@@ -354,7 +409,7 @@ class InputScope(PSBTScope):
 
 
 class OutputScope(PSBTScope):
-    def __init__(self, unknown: dict = {}, vout=None, compress=False):
+    def __init__(self, unknown: dict = {}, vout=None, compress=CompressMode.KEEP_ALL):
         self.compress = compress
         self.value = None
         self.script_pubkey = None
@@ -365,14 +420,20 @@ class OutputScope(PSBTScope):
         self.redeem_script = None
         self.witness_script = None
         self.bip32_derivations = OrderedDict()
+        self.taproot_bip32_derivations = OrderedDict()
+        self.taproot_internal_key = None
         self.parse_unknowns()
 
-    def clear_metadata(self):
+    def clear_metadata(self, compress=CompressMode.CLEAR_ALL):
         """Removes metadata like derivations, utxos etc except final or partial sigs"""
+        if compress == CompressMode.KEEP_ALL:
+            return
         self.unknown = {}
         self.redeem_script = None
         self.witness_script = None
         self.bip32_derivations = OrderedDict()
+        self.taproot_bip32_derivations = OrderedDict()
+        self.taproot_internal_key = None
 
     def update(self, other):
         self.value = other.value if other.value is not None else self.value
@@ -381,6 +442,8 @@ class OutputScope(PSBTScope):
         self.redeem_script = other.redeem_script or self.redeem_script
         self.witness_script = other.witness_script or self.witness_script
         self.bip32_derivations.update(other.bip32_derivations)
+        self.taproot_bip32_derivations.update(other.bip32_derivations)
+        self.taproot_internal_key = other.taproot_internal_key
 
     @property
     def vout(self):
@@ -390,7 +453,9 @@ class OutputScope(PSBTScope):
         # separator
         if len(k) == 0:
             return
+
         v = read_string(stream)
+
         # redeem script
         if k[0] == 0x00:
             if len(k) != 1:
@@ -414,15 +479,34 @@ class OutputScope(PSBTScope):
                 raise PSBTError("Duplicated derivation path")
             else:
                 self.bip32_derivations[pub] = DerivationPath.parse(v)
+
         elif k == b"\x03":
             self.value = int.from_bytes(v, 'little')
         elif k == b"\x04":
             self.script_pubkey = Script(v)
+
+        # PSBT_OUT_TAP_INTERNAL_KEY
+        elif k[0] == 0x05:
+            self.taproot_internal_key = ec.PublicKey.from_xonly(v)
+
+        # PSBT_OUT_TAP_BIP32_DERIVATION
+        elif k[0] == 0x07:
+            pub = ec.PublicKey.from_xonly(k[1:])
+            if pub not in self.taproot_bip32_derivations:
+                b = BytesIO(v)
+                num_leaf_hashes = compact.read_from(b)
+                leaf_hashes = [b.read(32) for i in range(num_leaf_hashes)]
+                if not all([len(leaf)==32 for leaf in leaf_hashes]):
+                    raise PSBTError("Invalid length of taproot leaf hashes")
+                der = DerivationPath.read_from(b)
+                self.taproot_bip32_derivations[pub] = (leaf_hashes, der)
+
         else:
             if k in self.unknown:
                 raise PSBTError("Duplicated key")
             self.unknown[k] = v
 
+
     def write_to(self, stream, skip_separator=False, version=None, **kwargs) -> int:
         r = 0
         if self.redeem_script is not None:
@@ -434,6 +518,7 @@ class OutputScope(PSBTScope):
         for pub in self.bip32_derivations:
             r += ser_string(stream, b"\x02" + pub.serialize())
             r += ser_string(stream, self.bip32_derivations[pub].serialize())
+
         if version == 2:
             if self.value is not None:
                 r += ser_string(stream, b"\x03")
@@ -442,6 +527,17 @@ class OutputScope(PSBTScope):
                 r += ser_string(stream, b"\x04")
                 r += self.script_pubkey.write_to(stream)
 
+        # PSBT_OUT_TAP_INTERNAL_KEY
+        if self.taproot_internal_key is not None:
+            r += ser_string(stream, b"\x05")
+            r += ser_string(stream, self.taproot_internal_key.xonly())
+
+        # PSBT_OUT_TAP_BIP32_DERIVATION
+        for pub in self.taproot_bip32_derivations:
+            r += ser_string(stream, b"\x07" + pub.xonly())
+            leaf_hashes, derivation = self.taproot_bip32_derivations[pub]
+            r += ser_string(stream, compact.to_bytes(len(leaf_hashes)) + b"".join(leaf_hashes) + derivation.serialize())
+
         # unknown
         for key in self.unknown:
             r += ser_string(stream, key)
@@ -557,7 +653,7 @@ class PSBT(EmbitBase):
         return r
 
     @classmethod
-    def from_base64(cls, b64, compress=False):
+    def from_base64(cls, b64, compress=CompressMode.KEEP_ALL):
         raw = a2b_base64(b64)
         return cls.parse(raw, compress=compress)
 
@@ -571,14 +667,14 @@ class PSBT(EmbitBase):
             return hexlify(self.serialize()).decode()
 
     @classmethod
-    def from_string(cls, s, compress=False):
+    def from_string(cls, s, compress=CompressMode.KEEP_ALL):
         if s.startswith(hexlify(cls.MAGIC).decode()):
             return cls.parse(unhexlify(s), compress=compress)
         else:
             return cls.from_base64(s, compress=compress)
 
     @classmethod
-    def read_from(cls, stream, compress=False):
+    def read_from(cls, stream, compress=CompressMode.KEEP_ALL):
         """
         Compress flag allows to load and verify non_witness_utxo
         without storing them in memory and save the utxo internally for signing.
@@ -741,32 +837,49 @@ class PSBT(EmbitBase):
                         counter += 1
                 # if we use HDKey
                 else:
-                    # TODO: add taproot derivation paths and scripts
+                    bip32_derivations = []
+                    for pub in inp.taproot_bip32_derivations:
+                        leaf_hashes, derivation = inp.taproot_bip32_derivations[pub]
+                        # TODO: also sign with leaf_hashes
+                        if derivation.fingerprint == fingerprint and len(leaf_hashes) == 0:
+                            bip32_derivations.append((pub, derivation))
+                    
+                    # "Legacy" support for workaround when BIP-371 Taproot psbt fields aren't available.
+                    # TODO: Remove this (and refactor above) when workaround has been phased out.
                     for pub in inp.bip32_derivations:
-                        # check if it is root key
-                        if inp.bip32_derivations[pub].fingerprint == fingerprint:
-                            der = inp.bip32_derivations[pub].derivation
-                            if hasattr(root, "origin"):
-                                # for descriptor key remove origin part
-                                if root.origin:
-                                    if root.origin.derivation != der[:len(root.origin.derivation)]:
-                                        continue
-                                    der = der[len(root.origin.derivation):]
-                                hdkey = root.key.derive(der)
-                            else:
-                                hdkey = root.derive(der)
-                            mypub = hdkey.key.get_public_key()
-                            if mypub != pub:
-                                raise PSBTError("Derivation path doesn't look right")
-                            pk = hdkey.taproot_tweak(b"")
-                            if pk.xonly() in sc.data:
-                                sig = pk.schnorr_sign(h)
-                                # sig plus sighash flag
-                                wit = sig.serialize()
-                                if inp_sighash != SIGHASH.DEFAULT:
-                                    wit += bytes([inp_sighash])
-                                inp.final_scriptwitness = Witness([wit])
-                                counter += 1
+                        derivation = inp.bip32_derivations[pub]
+                        if derivation.fingerprint == fingerprint:
+                            bip32_derivations.append((pub, derivation))
+
+                    for pub, derivation in bip32_derivations:
+                        der = derivation.derivation
+                        if hasattr(root, "origin"):
+                            # for descriptor key remove origin part
+                            if root.origin:
+                                if root.origin.derivation != der[:len(root.origin.derivation)]:
+                                    continue
+                                der = der[len(root.origin.derivation):]
+                            hdkey = root.key.derive(der)
+                        else:
+                            hdkey = root.derive(der)
+                        
+                        # Taproot BIP32 derivations use X-only pubkeys
+                        xonly_pub = hdkey.key.xonly()
+                        mypub = ec.PublicKey.from_xonly(xonly_pub)
+
+                        if mypub != pub:
+                            raise PSBTError("Derivation path doesn't look right")
+                        
+                        # TODO: Support signing for keys within leaves
+                        pk = hdkey.taproot_tweak(b"")
+                        if pk.xonly() in sc.data:
+                            sig = pk.schnorr_sign(h)
+                            # sig plus sighash flag
+                            wit = sig.serialize()
+                            if inp_sighash != SIGHASH.DEFAULT:
+                                wit += bytes([inp_sighash])
+                            inp.final_scriptwitness = Witness([wit])
+                            counter += 1
                 continue
 
             # if we have individual private key
@@ -800,4 +913,4 @@ class PSBT(EmbitBase):
                     # sig plus sighash flag
                     inp.partial_sigs[mypub] = sig.serialize() + bytes([inp_sighash])
                     counter += 1
-        return counter
+        return counter
\ No newline at end of file
